/* $Id: cmd-offl.c,v 1.3 1999/10/11 20:35:05 linus Exp $
 *
 * Copyright (C) 1999  Lysator Computer Club,
 *                     Linkoping University, Sweden
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 *  Author: Linus Tolke
 */

#include <config.h>

#include <sys/param.h>

#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <libintl.h>

#include <kom-errno.h>

#include <misc-parser.h>

#include "xoutput.h"
#include "copy.h"
#include "error.h"
#include "cache.h"
#include "quit.h"
#include "main.h"
#include "offline.h"

#define Export

/*
 * These operations are probably slow so let the user know what is going on 
 * by giving him a counter to look at.
 */
static int counter;
void trace_go(void)
{
    if (counter == 0)
	xprintf("        \r");
    else
    {
	xprintf("%d\r", counter);
    }
    counter ++;
    xflush(kom_outstream);
}
void trace_reset() { counter = 0; trace_go(); }

/*
 * Functions to keep track of what was recently fetched.
 * They implement a binary tree.
 */
#define NEWLY_FUNCTIONS_FOR_TYPE(type)					\
struct type ## _struct							\
{									\
    type			no;					\
    struct type ## _struct * 	left;					\
    struct type ## _struct * 	right;					\
};									\
									\
void									\
free_fetched_ ## type ## _struct(struct type ## _struct * p)		\
{									\
    if (p == NULL)							\
	return;								\
    free_fetched_ ## type ## _struct(p->left);				\
    free_fetched_ ## type ## _struct(p->right);				\
    free(p);								\
}									\
									\
int									\
insert_into_ ## type ## _list(type no, struct type ## _struct ** r)	\
{									\
    if (*r == NULL)							\
    {									\
	*r = (struct type ## _struct *)					\
		malloc(sizeof(struct type ## _struct));			\
									\
	(*r)->no = no;							\
	(*r)->left = NULL;						\
	(*r)->right = NULL;						\
									\
	return FALSE;							\
    }									\
									\
    if ((*r)->no > no)							\
	return insert_into_ ## type ## _list(no, &(*r)->left);		\
    else if ((*r)->no < no)						\
	return insert_into_ ## type ## _list(no, &(*r)->right);		\
    else								\
	return TRUE;							\
}									\
									\
int									\
newly_fetched_ ## type(type no)						\
{									\
    static struct type ## _struct * root = NULL;			\
									\
    if (no == 0)							\
    {									\
	free_fetched_ ## type ## _struct(root);				\
	root = NULL;							\
	return 0;							\
    }									\
									\
    return insert_into_ ## type ## _list(no, &root);			\
}

NEWLY_FUNCTIONS_FOR_TYPE(Conf_no);
NEWLY_FUNCTIONS_FOR_TYPE(Text_no);

/*
 * Compares times.
 * Returns true if the first time is before the second one else false.
 */
int
before(Time * first, Time * second)
{
#define CMP(kind)				\
    if (first->kind < second->kind)		\
	return TRUE;				\
    else if (first->kind > second->kind)	\
	return FALSE;

    CMP(tm_year);
    CMP(tm_mon);
    CMP(tm_mday);
    CMP(tm_hour);
    CMP(tm_min);
    CMP(tm_sec);
    return FALSE;
#undef CMP
}

/*
 * Returns true if the text with LOCAL NO is read in the MEMBERSHIP.
 */
int
already_read(Local_text_no lno, Membership * ms)
{
    int i;

    if (lno < ms->last_text_read)
	return TRUE;
    
    if (ms->read_texts == NULL)
	return FALSE;
    for (i = 0; i < ms->no_of_read; i++)
    {
	if (lno == ms->read_texts[i])
	    return TRUE;
    }
    return FALSE;
}

/*
 * Download all unread articles for later reading.
 *
 * Download is performed like this:
 * 1. Download a fresh membership. Save it.
 * 2. Download conf status of all the conferences in the membership.
 * 3. Download complete or missing part of the maps of the conferences that 
 *    have changed since last time.
 * 4. Step through the membership fetching texts and textstats that were not
 *    already there.
 * 5. For each textstat in the membership, refetch referenced textstats. 
 *    (This is probably a comment added to an existing text and must thus 
 *    be refetched.)
 * 6. For each textstat fetched in 4 or 5 fetch the confstat of the author if
 *    not already present in the database.
 *
 * Steps 3 - 6 can be done per conf. Step 5 and 6 can be done immediatly 
 * following the previous operation instead of saving them until last as
 * suggested by the list above.
 */
Export Success
cmd_download(String argument)
{
#define FUNCTION "cmd_download()"
    Success		retval;
    Membership_list	memb_list = EMPTY_MEMBERSHIP_LIST_i;
    unsigned short	mi;

    Conference		conf = EMPTY_CONFERENCE_i;

    Text_list		text_list = EMPTY_TEXT_LIST_i;
    Time		timestamp = NULL_TIME_i;

    String		textmass = EMPTY_STRING_i;
    Text_stat		textstat = EMPTY_TEXT_STAT_i;
    Text_stat		textstat2 = EMPTY_TEXT_STAT_i;

#define CLEAN_UP()				\
    do {					\
	release_membership_list(&memb_list);	\
	release_conf(&conf);			\
	release_text_list(&text_list);		\
	release_text_stat(&textstat);		\
	release_text_stat(&textstat2);		\
    } while (0)

    TOPLOOP_SETJMP();

    OFFLINE();

    trace_reset();
    trace_go();

    store_person_number(cpn);

    retval = kom_get_membership(cpn, 0, MAX_CONF_NO, TRUE, &memb_list);

    if (retval != OK)
    {
	fatal4 (CLIENT_UNEXPECTED_SERVER_ERROR,
		"get_membership(pers=%d, next_to_fetch=%d,chunk=%d)",
		cpn, 0, MAX_CONF_NO);
	/*NOTREACHED*/
    }

    trace_go();

    if (store_Membership_list(&memb_list) != OK)
    {
	RETURN(FAILURE);
    }

    newly_fetched_Text_no(0);	/* Empty the list. */
    newly_fetched_Conf_no(0);	/* Empty the list. */

    for (mi = 0; mi < memb_list.no_of_confs; mi++)
    {
	int confno = memb_list.confs[mi].conf_no;

	if (confno == 0)
	    continue;

	if (!newly_fetched_Conf_no(confno))
	{
	    trace_go();

	    retval = kom_get_conf_stat(confno, &conf);

	    if (retval != OK)
	    {
		continue;
	    }

	    trace_go();

	    if (store_Conference(confno, &conf) != OK)
	    {
		RETURN(FAILURE);
	    }
	}

	if (fetch_map(confno, &text_list, &timestamp) != OK
	    || before(&timestamp, &conf.last_written))
	{
	    /* We need to refetch this map. */
	    /* It would be better just to fetch the increment. */
	    int ti;

	    trace_go();

	    retval = kom_get_map(confno, 0, LONG_MAX, &text_list);

	    if (retval != OK)
	    {
		continue;
	    }

	    trace_go();

	    if (store_map(confno, &text_list, &conf.last_written) != OK)
	    {
		RETURN(FAILURE);
	    }

	    for (ti = 0; ti < text_list.no_of_texts; ti++)
	    {
		Text_no	tno;
		const Misc_info *	misc;
		Misc_info_group		misc_data;

		/* Skip over read texts. */
		if (already_read(text_list.first_local_no + ti, 
				 &memb_list.confs[mi]))
		{
		    continue;
		}

		tno = text_list.texts[ti];

		if (newly_fetched_Text_no(tno))
		    continue;

		if (tno == 0)
		    continue;

		/* Fetch the text data. */
		if (fetch_text(tno, &textmass) != OK)
		{
		    /* It is not already fetched. */

		    trace_go();

		    retval = kom_get_text(tno, 0, END_OF_STRING, &textmass);

		    if (retval != OK)
		    {
			continue;
		    }

		    trace_go();

		    if (store_text(tno, &textmass) != OK)
		    {
			RETURN(FAILURE);
		    }
		}

		/* Now fetch the text stat. */
		if (fetch_Text_stat(tno, &textstat) != OK)
		{
		    /* It is not already fetched. */

		    trace_go();

		    retval = kom_get_text_stat(tno, &textstat);

		    if (retval != OK)
		    {
			continue;
		    }

		    trace_go();

		    if (store_Text_stat(tno, &textstat) != OK)
		    {
			RETURN(FAILURE);
		    }
		}

		if (textstat.author == 0
		    || newly_fetched_Conf_no(textstat.author))
		{
		    /* This author is either secret or newly fetched. */
		}
		else
		{
		    trace_go();

		    retval = kom_get_conf_stat(textstat.author, &conf);

		    if (retval == OK)
		    {
			if (store_Conference(textstat.author, &conf) != OK)
			    RETURN(FAILURE);
		    }
		}

		for (misc = textstat.misc_items;
		     (misc_data = parse_next_misc(&misc,
						  textstat.misc_items +
						  textstat.no_of_misc)).type
			 != m_end_of_list;
		     )
		{
		    Text_no ntno = 0;
		    Conf_no ncno = 0;

		    switch (misc_data.type)
		    {
		    case m_footn_in:
			ntno = misc_data.footnoted_in;
			break;
		    case m_comm_in:
			ntno = misc_data.commented_in;
			break;
		    case m_footn_to:
			ntno = misc_data.footnote_to;
			break;
		    case m_comm_to:
			ntno = misc_data.comment_to;
			break;
		    case recpt:
			ncno = misc_data.recipient;
			break;
		    case cc_recpt:
			ncno = misc_data.cc_recipient;
			break;
		    case sent_by:
			ncno = misc_data.sender;
		    case loc_no:
		    case rec_time:
		    case sent_at:
			continue;
		    default:
			fprintf(stderr, "Unknown text-stat type: %lu\n",
				misc_data.type);
			continue;
		    }

		    if (ncno != 0)
		    {
			/* This was a misc item referencing a person.
			   Fetch that persons conf-stat, if not already
			   done. */
			if (fetch_Conference(ncno, &conf) != OK)
			{
			    trace_go();

			    newly_fetched_Conf_no(ncno);

			    retval = kom_get_conf_stat(ncno, &conf);

			    if (retval == OK)
			    {
				trace_go();

				if (store_Conference(ncno, &conf) != OK)
				    RETURN(FAILURE);
			    }
			}
		    }

		    if (ntno == 0)
			continue;

		    /* Some other text was referenced.
		       Fetch that text-stat.
		       */

		    if (newly_fetched_Text_no(ntno))
			continue;

		    trace_go();

		    retval = kom_get_text_stat(ntno, &textstat2);

		    if (retval != OK)
		    {
			continue;
		    }

		    /* Fetch the author. */

		    trace_go();

		    if (store_Text_stat(ntno, &textstat2) != OK)
		    {
			RETURN(FAILURE);
		    }

		    if (textstat2.author == 0
			|| fetch_Conference(textstat2.author, &conf) == OK)
		    {
			/* This author is either secret or newly fetched. */
		    }
		    else
		    {
			trace_go();

			newly_fetched_Conf_no(textstat2.author);

			retval = kom_get_conf_stat(textstat2.author, 
						   &conf);

			if (retval == OK)
			{
			    trace_go();

			    if (store_Conference(textstat2.author, &conf)
				!= OK)
				RETURN(FAILURE);
			}
		    }
		}
	    }
	}
    }

    RETURN(OK);
#undef CLEAN_UP
#undef FUNCTION
}
